"""
UP42 authentication mechanism and base requests functionality
"""
import json
from pathlib import Path
from typing import Dict, Optional, Union
from warnings import warn

import requests
import requests.exceptions
from oauthlib.oauth2 import BackendApplicationClient, MissingTokenError
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from tenacity import (
    Retrying,
    retry,
    retry_if_exception,
    retry_if_exception_type,
    stop_after_attempt,
    wait_fixed,
    wait_random_exponential,
)

from up42.utils import get_logger, get_up42_py_version

logger = get_logger(__name__)


class retry_if_401_invalid_token(retry_if_exception):
    """
    Custom tenacity error response that enables separate retry strategy for
    401 error (unauthorized response, unable decode JWT) due to invalid/timed out UP42 token.

    Adapted from https://github.com/alexwlchan/handling-http-429-with-tenacity
    """

    def __init__(self):
        def is_http_401_error(exception):
            return (
                isinstance(
                    exception,
                    (
                        requests.exceptions.HTTPError,
                        requests.exceptions.RequestException,
                    ),
                )
                and hasattr(exception.response, "status_code")  # check if response has status_code
                and exception.response.status_code == 401
            )

        super().__init__(predicate=is_http_401_error)


class retry_if_429_rate_limit(retry_if_exception):
    """
    Custom tenacity error response that enables separate retry strategy for
    429 HTTPError (too many requests) due to UP42 rate limitation.

    Adapted from https://github.com/alexwlchan/handling-http-429-with-tenacity
    """

    def __init__(self):
        def is_http_429_error(exception):
            return isinstance(exception, requests.exceptions.HTTPError) and exception.response.status_code == 429

        super().__init__(predicate=is_http_429_error)


class Auth:
    def __init__(
        self,
        cfg_file: Union[str, Path, None] = None,
        project_id: Optional[str] = None,
        project_api_key: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ):
        """
        The Auth class handles the authentication with UP42.

        Authenticate with UP42:
            https://sdk.up42.com/authentication/.

        Args:
            cfg_file: File path to the cfg.json with either
            {project_id: "...", project_api_key: "..."} or {username: "...", password: "..."}.
            project_id: The unique identifier of the project.
            project_api_key: The project-specific API key.
            username: The username for the UP42 account (email UP42 console).
            password: Password for the UP42 console login.
        """
        self.cfg_file = cfg_file
        self.workspace_id: Optional[str] = None
        self.token: Optional[str] = None

        local_vars = locals()
        credentials_filter = ["project_id", "project_api_key", "username", "password"]
        credentials_dict = {
            key: value for key, value in local_vars.items() if key in credentials_filter and value is not None
        }

        try:
            self.env: str = kwargs["env"]
        except KeyError:
            self.env = "com"
        try:
            self.authenticate: bool = kwargs["authenticate"]
        except KeyError:
            self.authenticate = True

        if project_id:
            self.project_id = project_id

        config = self._choose_credential_source(cfg_file, credentials_dict)

        if self.authenticate:
            self._set_credentials(source=config)
            self._get_token()
            self._get_workspace()
            logger.info("Authentication with UP42 successful!")

    def _choose_credential_source(self, cfg_file: Union[str, Path, None], kwargs: dict) -> dict:
        config = {}
        if cfg_file:
            try:
                with open(cfg_file, encoding="utf-8") as src:
                    config.update(json.load(src))
                logger.info("Got credentials from config file.")
            except FileNotFoundError as e:
                raise ValueError("Selected config file does not exist!") from e

            if set(config.keys()).intersection(set(kwargs.keys())):
                raise ValueError("Credentials must be provided either via config file or arguments, but not both")
        return config if config else kwargs

    def _set_credentials(self, source: dict):
        schemas = {
            "user": ("username", "password"),
            "project": ("project_id", "project_api_key"),
        }
        token_retrievers = {
            "user": self._get_token_account_based,
            "project": self._get_token_project_based,
        }

        self._credentials_id = None

        for schema, parameters in schemas.items():
            if set(parameters).issubset(source.keys()):
                if schema == "project":
                    warn(
                        "Project based authentication will be deprecated."
                        "Please follow authentication guidelines (/docs/authentication.md)."
                    )
                self._credentials_id = source[parameters[0]]
                self._credentials_key = source[parameters[1]]
                self._get_token = token_retrievers[schema]
        if self._credentials_id is None:
            raise ValueError("No credentials provided either via config file or arguments.")

    def _endpoint(self) -> str:
        """Gets the endpoint."""
        return f"https://api.up42.{self.env}"

    def _get_token_account_based(self):
        """Account based authentication via username and password."""
        try:
            req_headers = {
                "Content-Type": "application/x-www-form-urlencoded",
            }
            req_body = {
                "grant_type": "password",
                "username": self._credentials_id,
                "password": self._credentials_key,
            }
            token_response = requests.post(
                url=self._endpoint() + "/oauth/token",
                data=req_body,
                headers=req_headers,
                timeout=120,
            )
            if token_response.status_code != 200:
                raise ValueError(
                    f"Authentication failed with status code {token_response.status_code}."
                    "Check the provided credentials."
                )
        except requests.exceptions.RequestException as err:
            raise ValueError(
                "Authentication failed due to a network error. Check the provided credentials and network connectivity."
            ) from err
        self.token = token_response.json()["data"]["accessToken"]

    def _get_token_project_based(self):
        """Project specific authentication via project id and project api key."""
        try:
            client = BackendApplicationClient(client_id=self._credentials_id, client_secret=self._credentials_key)
            auth = HTTPBasicAuth(self._credentials_id, self._credentials_key)
            get_token_session = OAuth2Session(client=client)
            token_response = get_token_session.fetch_token(token_url=self._endpoint() + "/oauth/token", auth=auth)
        except MissingTokenError as err:
            raise ValueError("Authentication was not successful, check the provided project credentials.") from err

        self.token = token_response["data"]["accessToken"]

    def _get_workspace(self) -> None:
        """Get user id belonging to authenticated account."""
        url = f"https://api.up42.{self.env}/users/me"
        resp = self._request("GET", url)
        self.workspace_id = resp["data"]["id"]  # type: ignore

    @staticmethod
    def _generate_headers(token: str) -> Dict[str, str]:
        version = get_up42_py_version()
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
            "cache-control": "no-cache",
            "User-Agent": f"up42-py/{version} (https://github.com/up42/up42-py)",
        }
        return headers

    # pylint: disable=dangerous-default-value
    @retry(  # type: ignore
        retry=retry_if_429_rate_limit(),
        wait=wait_random_exponential(multiplier=0.5, max=180),
        reraise=True,
    )
    def _request_helper(
        self,
        request_type: str,
        url: str,
        data: dict = {},
        querystring: dict = {},
    ) -> requests.Response:
        """
        Helper function for the request, running the actual request with the correct headers.

        Args:
            request_type: 'GET', 'POST', 'PUT', 'PATCH', 'DELETE'
            url: The requests url.
            data: The payload, e.g. dictionary with job parameters etc.
            querystring: The querystring.

        Returns:
            The request response.
        """
        headers = self._generate_headers(self.token)  # type: ignore
        if querystring == {}:
            response: requests.Response = requests.request(
                method=request_type,
                url=url,
                data=json.dumps(data),
                headers=headers,
                timeout=120,
            )
        else:
            response = requests.request(
                method=request_type,
                url=url,
                data=json.dumps(data),
                headers=headers,
                params=querystring,
                timeout=120,
            )
        logger.debug(response)
        logger.debug(data)
        response.raise_for_status()
        return response

    def _request(
        self,
        request_type: str,
        url: str,
        data: Union[dict, list] = {},
        querystring: dict = {},
        return_text: bool = True,
    ):  # Union[str, dict, requests.Response]:
        """
        Handles retrying the request and automatically retries and gets a new token if
        the old is invalid.

        Retry is enabled by default, can be set to False as kwargs of Auth.

        In addition to this retry mechanic, 429-errors (too many requests) are retried
        more extensively in _request_helper.

        Args:
            request_type: 'GET', 'POST', 'PUT', 'PATCH', 'DELETE'
            url: The url to request.
            data: The payload, e.g. dictionary with job parameters etc.
            querystring: The querystring.
            return_text: If true returns response text/json, false returns response.
            retry: If False, after 5 minutes and invalid token will return 401
                errors.

        Returns:
            The API response.
        """
        retryer_token = Retrying(
            stop=stop_after_attempt(2),  # Original attempt + one retry
            wait=wait_fixed(0.5),
            retry=(retry_if_401_invalid_token() | retry_if_exception_type(requests.exceptions.ConnectionError)),
            after=lambda retry_state: self._get_token(),  # type:ignore
            reraise=True,
            # after final failed attempt, raises last attempt's exception instead of RetryError.
        )

        try:
            response: requests.Response = retryer_token(self._request_helper, request_type, url, data, querystring)

        # There are two UP42 API versions:
        # v1 endpoints give response format {"data": ..., "error": ...}   data e.g. dict or list.  error str or dict
        # or None (if no error).
        # v1 always gives response, the error is indicated by the error key.
        # v2 endpoints follows RFC 7807: {"title":..., "status": 404} Optional "detail" and "type" keys.
        # v2 either gives above positive response, or fails with httperror (then check error.json() for the above
        # fields)

        except requests.exceptions.RequestException as err:  # Base error class
            # Raising the original `err` error would not surface the relevant error message (contained in API response)
            err_message = err.response.json()
            logger.error(f"Error {err_message}")
            raise requests.exceptions.RequestException(err_message) from err

        # Handle response text.
        if return_text:
            try:
                response_text = json.loads(response.text)
            except json.JSONDecodeError:  # e.g. JobTask logs are str format.
                response_text = response.text

            # Handle api error messages here before handling it in every single function.
            try:
                if response_text["error"] is not None and response_text["data"] is None:
                    raise ValueError(response_text["error"])
                return response_text
            except (
                KeyError,
                TypeError,
            ):  # Catalog search, JobTask logs etc. does not have the usual {"data":"",
                # "error":""} format.
                return response_text

        else:  # E.g. for DELETE
            return response



from datetime import datetime
from enum import Enum
from typing import List, Optional, Union
from urllib.parse import urlencode, urljoin
from warnings import warn

from geojson import Feature, FeatureCollection
from geopandas import GeoDataFrame
from shapely.geometry import Polygon

from up42.asset import Asset
from up42.asset_searcher import AssetSearchParams, query_paginated_endpoints, search_assets
from up42.auth import Auth
from up42.order import Order
from up42.stac_client import PySTACAuthClient
from up42.utils import format_time, get_logger

logger = get_logger(__name__)


class AllowedStatuses(Enum):
    CREATED = "CREATED"
    BEING_PLACED = "BEING_PLACED"
    PLACED = "PLACED"
    PLACEMENT_FAILED = "PLACEMENT_FAILED"
    DELIVERY_INITIALIZATION_FAILED = "DELIVERY_INITIALIZATION_FAILED"
    BEING_FULFILLED = "BEING_FULFILLED"
    DOWNLOAD_FAILED = "DOWNLOAD_FAILED"
    DOWNLOADED = "DOWNLOADED"
    FULFILLED = "FULFILLED"
    FAILED_PERMANENTLY = "FAILED_PERMANENTLY"


class Storage:
    """
    The Storage class enables access to the UP42 storage. You can list
    your assets and orders within an UP42 workspace.

    Use the storage:
    ```python
    storage = up42.initialize_storage()
    ```
    """

    def __init__(self, auth: Auth):
        self.auth = auth
        self.workspace_id = auth.workspace_id

    def __repr__(self):
        env = ", env: dev" if self.auth.env == "dev" else ""
        return f"Storage(workspace_id: {self.workspace_id}{env})"

    @property
    def pystac_client(self):
        url = f"{self.auth._endpoint()}/v2/assets/stac"
        pystac_client_auth = PySTACAuthClient(auth=self.auth).open(url=url)
        return pystac_client_auth

    def _query_paginated_stac_search(
        self,
        url: str,
        stac_search_parameters: dict,
    ) -> list:
        """
        Helper to fetch list of items in paginated stac search endpoind, e.g. stac search assets.

        Args:
            url (str): The base url for paginated endpoint.
            stac_search_parameters (dict): the parameters required for stac search

        Returns:
            List of storage STAC results features.
        """
        response_features: list = []
        response_features_limit = stac_search_parameters["limit"]
        while len(response_features) < response_features_limit:
            stac_results = self.auth._request(request_type="POST", url=url, data=stac_search_parameters)
            response_features.extend(stac_results["features"])
            token_list = [link["body"]["token"] for link in stac_results["links"] if link["rel"] == "next"]
            if token_list:
                stac_search_parameters["token"] = token_list[0]
            else:
                break
        return response_features

    def get_assets(
        self,
        created_after: Optional[Union[str, datetime]] = None,
        created_before: Optional[Union[str, datetime]] = None,
        acquired_after: Optional[Union[str, datetime]] = None,
        acquired_before: Optional[Union[str, datetime]] = None,
        geometry: Optional[Union[dict, Feature, FeatureCollection, list, GeoDataFrame, Polygon]] = None,
        workspace_id: Optional[str] = None,
        collection_names: List[str] = None,
        producer_names: List[str] = None,
        tags: List[str] = None,
        sources: List[str] = None,
        search: str = None,
        custom_filter: dict = None,
        limit: Optional[int] = None,
        sortby: str = "createdAt",
        descending: bool = True,
        return_json: bool = False,
    ) -> Union[List[Asset], List[dict]]:
        """
        Gets a list of assets in storage as [Asset](https://sdk.up42.com/structure/#asset) objects or in JSON format.

        Args:
            created_after: Search for assets created after the specified timestamp, in `"YYYY-MM-DD"` format.
            created_before: Search for assets created before the specified timestamp, in `"YYYY-MM-DD"` format.
            acquired_after: Deprecated filter.
            acquired_before: Deprecated filter.
            geometry: Deprecated filter.
            workspace_id: Search by the workspace ID.
            collection_names: Search for assets from any of the provided geospatial collections.
            producer_names: Search for assets from any of the provided producers.
            tags: Search for assets with any of the provided tags.
            sources: Search for assets from any of the provided sources.\
                The allowed values: `"ARCHIVE"`, `"TASKING"`, `"USER"`.
            search: Search for assets that contain the provided search query in their name, title, or order ID.
            custom_filter: Deprecated filter.
            limit: The number of results on a results page.
            sortby: The property to sort by.
            descending: The sorting order: <ul><li>`true` — descending</li><li>`false` — ascending</li></ul>
            return_json: If `true`, returns a JSON dictionary.\
                If `false`, returns a list of [Asset](https://sdk.up42.com/structure/#functionality_1) objects.

        Returns:
            A list of Asset objects.
        """
        params: AssetSearchParams = {
            "createdAfter": created_after and format_time(created_after),
            "createdBefore": created_before and format_time(created_before),
            "workspaceId": workspace_id,
            "collectionNames": collection_names,
            "producerNames": producer_names,
            "tags": tags,
            "sources": sources,
            "search": search,
        }
        assets_json = search_assets(
            self.auth,
            params=params,
            limit=limit,
            sortby=sortby,
            descending=descending,
        )

        if (
            acquired_before is not None
            or acquired_after is not None
            or geometry is not None
            or custom_filter is not None
        ):
            warn(
                "Search for geometry, acquired_before, acquired_after and custom_filter has been deprecated."
                "Use the PySTAC client for STAC queries: https://sdk.up42.com/notebooks/stac-example/#pystac",
                DeprecationWarning,
                stacklevel=2,
            )
            raise ValueError(
                "Search for geometry, acquired_before, acquired_after and custom_filter is no longer supported."
            )

        if return_json:
            return assets_json
        else:
            return [Asset(self.auth, asset_id=asset_json["id"], asset_info=asset_json) for asset_json in assets_json]

    def get_orders(
        self,
        workspace_orders: bool = True,
        return_json: bool = False,
        limit: Optional[int] = None,
        sortby: str = "createdAt",
        descending: bool = True,
        order_type: Optional[str] = None,
        statuses: Optional[List[AllowedStatuses]] = None,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> Union[List[Order], List[dict]]:
        """
        Gets all orders in the account/workspace as Order objects or JSON.

        Args:
            workspace_orders: If set to True, only returns workspace orders. Otherwise, returns all account orders.
            return_json: If set to True, returns JSON object.
            limit: Optional, only return n first assets by sorting criteria and order.
                Optimal to select if your workspace contains many assets.
            sortby: The sorting criteria, one of "createdAt", "updatedAt", "status", "dataProvider", "type".
            descending: The sorting order, True for descending (default), False for ascending.
            order_type: Can be either "TASKING" or "ARCHIVE". Pass this param to filter orders based on order_type.
            statuses: Search for orders with any of the statuses provided.
            name: Search for orders that contain this string in their name.
            tags: Search for orders with any of the provided tags.

        Returns:
            Order objects in the workspace or alternatively JSON info of the orders.
        """
        allowed_statuses = {entry.value for entry in AllowedStatuses}

        allowed_sorting_criteria = {
            "createdAt",
            "updatedAt",
            "type",
            "status",
        }
        if sortby not in allowed_sorting_criteria:
            raise ValueError(f"sortby parameter must be one of {allowed_sorting_criteria}!")
        sort = f"{sortby},{'desc' if descending else 'asc'}"
        base_url = f"{self.auth._endpoint()}/v2/orders"

        params = {
            "sort": sort,
            "workspaceId": self.workspace_id if workspace_orders else None,
            "displayName": name,
            "type": order_type if order_type in ["TASKING", "ARCHIVE"] else None,
            "tags": tags,
            "status": set(statuses) & allowed_statuses if statuses else None,
        }
        params = {k: v for k, v in params.items() if v is not None}

        if statuses is not None and len(statuses) > len(params["status"]):
            logger.info(
                "statuses not included in allowed_statuses"
                f"{set(statuses).difference(allowed_statuses)} were ignored."
            )

        url = urljoin(base_url, "?" + urlencode(params, doseq=True, safe=""))

        orders_json = query_paginated_endpoints(auth=self.auth, url=url, limit=limit)
        logger_message = f"Got {len(orders_json)} orders" + (
            f" for workspace {self.workspace_id}." if workspace_orders else ""
        )
        logger.info(logger_message)

        if return_json:
            return orders_json
        else:
            return [Order(self.auth, order_id=order_json["id"], order_info=order_json) for order_json in orders_json]



import math
from datetime import datetime
from typing import Any, List, Optional, TypedDict, Union
from urllib.parse import urlencode, urljoin

from up42.utils import get_logger

logger = get_logger(__name__)


def query_paginated_endpoints(auth, url: str, limit: Optional[int] = None, size: int = 50) -> List[dict]:
    """
    Helper to fetch list of items in paginated endpoint, e.g. assets, orders.

    Args:
        url (str): The base url for paginated endpoint.
        limit: Return n first elements sorted by date of creation, optional.
        size: Default number of results per pagination page. Tradeoff of number
            of results per page and API response time to query one page. Default 50.

    Returns:
        List[dict]: List of all paginated items.
    """
    url = url + f"&size={size}"

    first_page_response = auth._request(request_type="GET", url=url)
    if "data" in first_page_response:  # UP42 API v2 convention without data key, but still in e.g. get order
        # endpoint
        first_page_response = first_page_response["data"]
    num_pages = first_page_response["totalPages"]
    num_elements = first_page_response["totalElements"]
    results_list = first_page_response["content"]

    if limit is None:
        # Also covers single page (without limit)
        num_pages_to_query = num_pages
    elif limit <= size:
        return results_list[:limit]
    else:
        # Also covers single page (with limit)
        num_pages_to_query = math.ceil(min(limit, num_elements) / size)

    for page in range(1, num_pages_to_query):
        response_json = auth._request(request_type="GET", url=url + f"&page={page}")
        if "data" in response_json:
            response_json = response_json["data"]
        results_list += response_json["content"]
    return results_list[:limit]


class AssetSearchParams(TypedDict, total=False):
    createdAfter: Optional[Union[str, datetime]]
    createdBefore: Optional[Union[str, datetime]]
    workspaceId: Optional[str]
    collectionNames: Optional[List[str]]
    producerNames: Optional[List[str]]
    tags: Optional[List[str]]
    sources: Optional[List[str]]
    search: Optional[str]


def search_assets(
    auth,
    params: AssetSearchParams,
    limit: Optional[int] = None,
    sortby: str = "createdAt",
    descending: bool = True,
) -> List[dict]:
    """
    Get a list of assets based on specified search parameters.

    Args:
        auth: An instance of the authentication class.
        params: A dictionary containing search parameters defined by the `AssetSearchParams` TypedDict.
        limit: The number of results on a results page (default is `None`).
        sortby: The property to sort the results by (default is "createdAt").
        descending: The sorting order, where `True` is descending and `False` is ascending (default is `True`).

    Returns:
        A list of assets metadata as dictionaries.
    """
    sort = f"{sortby},{'desc' if descending else 'asc'}"
    request_params: dict[str, Any] = {"sort": sort}
    request_params.update({key: value for key, value in params.items() if value is not None})
    base_url = f"{auth._endpoint()}/v2/assets"
    url = urljoin(base_url, "?" + urlencode(request_params, doseq=True, safe=""))
    assets_json = query_paginated_endpoints(auth, url=url, limit=limit)

    if "workspace_id" in request_params:
        logger.info(f"Queried {len(assets_json)} assets for workspace {auth.workspace_id}.")
    else:
        logger.info(f"Queried {len(assets_json)} assets from all workspaces in account.")
    return assets_json



"""
    `up42` is the base library module imported to Python. It provides the elementary
    functionality that is not bound to a specific class of the UP42 structure.
    From `up42` you can also initialize other classes, e.g. for using the catalog, storage etc.

    To import the UP42 library:
    ```python
    import up42
    ```

    Authenticate with UP42:
        https://sdk.up42.com/authentication/.

    To initialize any lower level functionality use e.g.
    ```python
    catalog = up42.initialize_catalog()
    ```
    ```python
    project = up42.initialize_project(project_id="your-project-ID")
    ```
"""

from pathlib import Path

from up42.asset import Asset
from up42.auth import Auth
from up42.catalog import Catalog
from up42.initialization import *  # defined by __all__
from up42.job import Job
from up42.jobcollection import JobCollection
from up42.jobtask import JobTask
from up42.main import *  # defined by __all__
from up42.order import Order
from up42.project import Project
from up42.storage import Storage
from up42.tasking import Tasking
from up42.tools import get_example_aoi, read_vector_file
from up42.utils import get_up42_py_version
from up42.viztools import draw_aoi
from up42.webhooks import Webhook
from up42.workflow import Workflow

__version__ = get_up42_py_version()



from pathlib import Path
from typing import List, Optional, Tuple, Union

import pystac
from pystac_client import Client, ItemSearch

from up42.auth import Auth
from up42.stac_client import PySTACAuthClient
from up42.utils import download_from_gcs_unpack, download_gcs_not_unpack, get_filename, get_logger

logger = get_logger(__name__)

MAX_ITEM = 50
LIMIT = 50


class Asset:
    """
    The Asset class enables access to the UP42 assets in the storage. Assets are results
    of orders or results of jobs with download blocks.

    Use an existing asset:
    ```python
    asset = up42.initialize_asset(asset_id="8c2dfb4d-bd35-435f-8667-48aea0dce2da")
    ```
    """

    def __init__(self, auth: Auth, asset_id: str, asset_info: Optional[dict] = None):
        self.auth = auth
        self.asset_id = asset_id
        self.results: Union[List[str], None] = None
        if asset_info is not None:
            self._info = asset_info
        else:
            self._info = self.info

    def __repr__(self):
        representation = (
            f"Asset(name: {self._info['name']}, asset_id: {self.asset_id}, createdAt: {self._info['createdAt']}, "
            f"size: {self._info['size']})"
        )
        if "source" in self._info:
            representation += f", source: {self._info['source']}"
        if "contentType" in self._info:
            representation += f", contentType: {self._info['contentType']}"
        return representation

    @property
    def info(self) -> dict:
        """
        Gets and updates the asset metadata information.
        """
        url = f"{self.auth._endpoint()}/v2/assets/{self.asset_id}/metadata"
        response_json = self.auth._request(request_type="GET", url=url)
        self._info = response_json
        return self._info

    @property
    def _stac_search(self) -> Tuple[Client, ItemSearch]:
        url = f"{self.auth._endpoint()}/v2/assets/stac"
        pystac_client_aux = PySTACAuthClient(auth=self.auth).open(url=url)
        stac_search_parameters = {
            "max_items": MAX_ITEM,
            "limit": LIMIT,
            "filter": {
                "op": "=",
                "args": [
                    {"property": "asset_id"},
                    self.asset_id,
                ],
            },
        }
        pystac_asset_search = pystac_client_aux.search(filter=stac_search_parameters)
        return (pystac_client_aux, pystac_asset_search)

    @property
    def stac_info(self) -> Optional[pystac.Collection]:
        """
        Gets the storage STAC information for the asset as a FeatureCollection.
        One asset can contain multiple STAC items (e.g. the PAN and multispectral images).
        """
        pystac_client_aux, pystac_asset_search = self._stac_search
        resulting_item = pystac_asset_search.item_collection()
        if resulting_item is None:
            raise ValueError(f"No STAC metadata information available for this asset {self.asset_id}")
        collection_id = resulting_item[0].collection_id
        return pystac_client_aux.get_collection(collection_id)

    @property
    def stac_items(self) -> pystac.ItemCollection:
        """Returns the stac items from an UP42 asset STAC representation."""
        try:
            _, pystac_asset_search = self._stac_search
            resulting_items = pystac_asset_search.item_collection()
            return resulting_items
        except Exception as exc:
            raise ValueError(f"No STAC metadata information available for this asset {self.asset_id}") from exc

    def update_metadata(self, title: str = None, tags: List[str] = None, **kwargs) -> dict:
        """
        Update the metadata of the asset.

        Args:
            title: The title string to be assigned to the asset.
            tags: A list of tag strings to be assigned to the asset.

        Returns:
            The updated asset metadata information
        """
        url = f"{self.auth._endpoint()}/v2/assets/{self.asset_id}/metadata"
        body_update = {"title": title, "tags": tags, **kwargs}
        response_json = self.auth._request(request_type="POST", url=url, data=body_update)
        self._info = response_json
        return self._info

    def _get_download_url(self, stac_asset_id: Optional[str] = None, request_type: str = "POST") -> str:
        if stac_asset_id is None:
            url = f"{self.auth._endpoint()}/v2/assets/{self.asset_id}/download-url"
        else:
            url = f"{self.auth._endpoint()}/v2/assets/{stac_asset_id}/download-url"
        response_json = self.auth._request(request_type=request_type, url=url)
        download_url = response_json["url"]
        return download_url

    def get_stac_asset_url(self, stac_asset: pystac.Asset):
        """
        Returns the signed URL for the STAC Asset.
        Args:
            stac_asset: pystac Asset object.

        Returns:
            Signed URL for the STAC Asset.
        """
        stac_asset_id = stac_asset.href.split("/")[-1]
        # so we can utilize all functionalities of Auth class
        return self._get_download_url(stac_asset_id=stac_asset_id)

    def download(
        self,
        output_directory: Union[str, Path, None] = None,
        unpacking: bool = True,
    ) -> List[str]:
        """
        Downloads the asset. Unpacking the downloaded file will happen as default.

        Args:
            output_directory: The file output directory, defaults to the current working
                directory.
            unpacking: By default the download TGZ/TAR or ZIP archive file will be unpacked.

        Returns:
            List of the downloaded asset filepaths.
        """
        logger.info(f"Downloading asset {self.asset_id}")

        if output_directory is None:
            output_directory = Path.cwd() / f"asset_{self.asset_id}"
        else:
            output_directory = Path(output_directory)
        output_directory.mkdir(parents=True, exist_ok=True)
        logger.info(f"Download directory: {str(output_directory)}")

        download_url = self._get_download_url()
        if unpacking:
            out_filepaths = download_from_gcs_unpack(
                download_url=download_url,
                output_directory=output_directory,
            )
        else:
            out_filepaths = download_gcs_not_unpack(
                download_url=download_url,
                output_directory=output_directory,
            )

        self.results = out_filepaths
        return out_filepaths

    def download_stac_asset(
        self,
        stac_asset: pystac.Asset,
        output_directory: Union[str, Path, None] = None,
    ) -> Path:
        """
        Downloads a STAC asset to a specified output directory.

        Args:
            self (object): The instance of the class calling this method.
            stac_asset (pystac.Asset): The STAC asset to be downloaded.
            output_directory (Union[str, Path, None], optional): The directory where the asset
                will be downloaded. If not provided, a default directory structure will be used.

        Returns:
            Path: The path to the downloaded asset file.

        Raises:
            requests.exceptions.HTTPError: If there is an HTTP error during the download.

        """
        logger.info(f"Downloading STAC asset {stac_asset.title}")
        if output_directory is None:
            output_directory = Path.cwd() / f"asset_{self.asset_id}/{stac_asset.title}"
        else:
            output_directory = Path(output_directory)
        output_directory.mkdir(parents=True, exist_ok=True)
        logger.info(f"Download directory: {str(output_directory)}")
        download_url = self.get_stac_asset_url(stac_asset=stac_asset)
        file_name = get_filename(download_url, default_filename="stac_asset")
        out_file_path = output_directory / file_name
        download_gcs_not_unpack(download_url, output_directory)
        return out_file_path



from pathlib import Path
from typing import Any, Callable, Dict, List, Union

import geojson
from geojson import FeatureCollection

from up42.auth import Auth
from up42.job import Job
from up42.utils import get_logger
from up42.viztools import VizTools

logger = get_logger(__name__)


class JobCollection(VizTools):
    """
    The JobCollection class provides facilities for handling and downloading
    multiple jobs results as one object.

    A jobcollection is created as the result of e.g. running multiple jobs in parallel:
    ```python
    jobcollection = workflow.run_jobs_parallel()
    ```

    Initialize a jobcollection from existing jobs:
    ```python
    jobcollection = up42.initialize_jobcollection(
        job_ids=["12345", "6789"],
        project_id="uz92-8uo0-4dc9-ab1d-06d7ec1a5321"
    )
    ```
    """

    def __init__(self, auth: Auth, project_id: str, jobs: List[Job]):
        self.auth = auth
        self.project_id = project_id
        self.jobs = jobs
        if jobs is not None:
            self.jobs_id = [job.job_id for job in jobs]
        else:
            self.jobs_id = None

    def __repr__(self):
        return f"JobCollection(len: {len(self.jobs)}, jobs: {self.jobs}"

    def __getitem__(self, index: int) -> Job:
        return self.jobs[index]

    def __iter__(self):
        for job in self.jobs:
            yield job

    @property
    def info(self) -> Dict[str, dict]:
        """
        Gets and updates the metadata information for each job in the jobcollection,
            dictionary of job_id : job_information.
        """
        return self.apply(lambda job: job.info, only_succeeded=False)

    @property
    def status(self) -> Dict[str, str]:
        """
        Gets the status for each job in the jobcollection, a dictionary with
        job_id : job status.
        """
        return self.apply(lambda job: job.status, only_succeeded=False)

    def apply(self, worker: Callable, only_succeeded: bool = True, **kwargs) -> Dict[str, Any]:
        """
        Helper function to apply `worker` on all jobs in the collection.
        `worker` needs to accept `Job` as first argument. For example, a
        lambda function that returns the job info:
        ```python
        self.apply(lambda job: job.info)
        ```

        Args:
            worker: A function to apply on all jobs in the collection.
            only_succeeded: Only apply to succeeded jobs (default is `True`).
            kwargs: additional keyword arguments to pass to `worker`.
        Returns:
            Dictionary where the key is the job id and the value the return
            of `worker`.
        """
        if not self.jobs:
            raise ValueError("This is an empty JobCollection. Cannot apply over an empty job list.")

        out_dict = {}
        for job in self.jobs:
            if only_succeeded:
                if job.is_succeeded:
                    out_dict[job.job_id] = worker(job, **kwargs)
            else:
                out_dict[job.job_id] = worker(job, **kwargs)

        if not out_dict:
            raise ValueError("All jobs have failed! Cannot apply over an empty succeeded job list.")

        return out_dict

    # TODO: Add method to get logs of failed jobs

    def download_results(
        self,
        output_directory: Union[str, Path, None] = None,
        merge: bool = True,
        unpacking: bool = True,
    ) -> Dict[str, List[str]]:
        """
        Downloads the job results. The final results are individually downloaded
        and by default a merged data.json is generated with all the results in a single
        feature collection. Unpacking the final will happen as default.
        Args:
            output_directory: The file output directory, defaults to the current working
                directory.
            merge: Wether to generate a merged data.json with all results.
            unpacking: By default the final result which is in TAR archive format will be unpacked.

        Returns:
            Dict of the job_ids and jobs' downloaded results filepaths. In addition,
            an additional key merged_result is added with the path to the merged
            data.json.
        """
        if output_directory is None:
            output_directory = Path.cwd() / f"project_{self.project_id}"
        else:
            output_directory = Path(output_directory)

        def download_results_worker(job, output_directory, unpacking):
            out_dir = output_directory / f"job_{job.job_id}"
            out_filepaths_job = job.download_results(output_directory=out_dir, unpacking=unpacking)
            return out_filepaths_job

        out_filepaths = self.apply(
            download_results_worker,
            output_directory=output_directory,
            unpacking=unpacking,
        )

        if merge:
            merged_data_json = output_directory / "data.json"
            with open(merged_data_json, "w") as dst:
                out_features = []
                for job_id in out_filepaths:
                    all_files = out_filepaths[job_id]
                    data_json = [d for d in all_files if Path(d).name == "data.json"][0]
                    with open(data_json) as src:
                        data_json_fc = geojson.load(src)
                        for feat in data_json_fc.features:
                            feat.properties["job_id"] = job_id
                            try:
                                feat.properties["up42.data_path"] = f"job_{job_id}/{feat.properties['up42.data_path']}"
                            except KeyError:
                                logger.warning("data.json does not contain up42.data_path, skipping...")
                            out_features.append(feat)
                geojson.dump(FeatureCollection(out_features), dst)

            out_filepaths["merged_result"] = [str(merged_data_json)]

        self.results = out_filepaths
        return out_filepaths
